/* Copyright (c) 2004 Marijn Haverbeke
 *
 * This software is provided 'as-is', without any express or implied
 * warranty. In no event will the authors be held liable for any
 * damages arising from the use of this software.
 *
 * Permission is granted to anyone to use this software for any
 * purpose, including commercial applications, and to alter it and
 * redistribute it freely, subject to the following restrictions:
 *
 * 1. The origin of this software must not be misrepresented; you must
 *    not claim that you wrote the original software. If you use this
 *    software in a product, an acknowledgment in the product
 *    documentation would be appreciated but is not required.
 *
 * 2. Altered source versions must be plainly marked as such, and must
 *    not be misrepresented as being the original software.
 *
 * 3. This notice may not be removed or altered from any source
 *    distribution.
 *
 * Marijn Haverbeke
 * marijn(at)haverbeke.nl
 */

// The canvas is a window taking up most of the space of the main
// frame, showing the puzzle pieces and handling manipulating them.

#include <algorithm>
#include "canvas.hpp"
#include "jigsaw.hpp"
#include "ids.hpp"

namespace {
  // 'Sweep' a rectangle along a vector (well not really, but grow it
  // to contain the result of that sweeping)
  void Grow_Rectangle(wxRect& rect, const wxPoint& move)
  {
    if (move.x < 0)
    {
      rect.x += move.x;
      rect.width -= move.x;
    }
    else
    {
      rect.width += move.x;
    }
    if (move.y < 0)
    {
      rect.y += move.y;
      rect.height -= move.y;
    }
    else
    {
      rect.height += move.y;
    }
  }
}

// Event table for Canvas. Mostly mouse events, but also painting and
// resizing.
BEGIN_EVENT_TABLE(Canvas, wxWindow)
  EVT_PAINT(Canvas::Paint)
  EVT_SIZE(Canvas::Resize)
  EVT_LEFT_DOWN(Canvas::Mouse_Down)
  EVT_LEFT_UP(Canvas::Mouse_Up)
  EVT_RIGHT_DOWN(Canvas::Start_Clear_Area)
  EVT_RIGHT_UP(Canvas::Clear_Area)
  EVT_MOTION(Canvas::Mouse_Move)
  END_EVENT_TABLE();

Canvas::Canvas(wxWindow* parent)
  : wxWindow(parent, id_canvas, wxPoint(0, 0), wxDefaultSize),
    _game(NULL),
    _dragging(NULL),
    // -1 means 'not currently dragging an area to clear'
    _clear_drag_start(-1, -1),
    // The back buffer - contains an image of the background and the
    // pieces that are not being dragged around.
    _back_buffer(GetClientSize().GetWidth(), GetClientSize().GetHeight()),
    // The background image
    _background(NULL),
    // The background color
    _back_color(*wxBLACK),
    // Is the puzzle completed?
    _completed(false),
    // Font to show "Completed!" in
    _completed_font(18, wxDECORATIVE, wxNORMAL, wxBOLD)
{
  Repaint_Back_Buffer();
}

Canvas::~Canvas()
{
  delete _background;
}

// Start a new game
void Canvas::Set_Game(Jigsaw* game)
{
  _game = game;
  // Reset state variables
  _dragging = NULL;
  _completed = false;

  // Scatter the pieces over the available area
  if (_game != NULL)
  {
    wxSize canvas_size = GetClientSize();
    _game->Scatter_Pieces(canvas_size.GetWidth(), canvas_size.GetHeight());
  }

  // Repaint
  Repaint_Back_Buffer();
  Refresh(false);
}

// The windowing system tells us to repaint
void Canvas::Paint(wxPaintEvent& event)
{
  wxPaintDC dc(this);
  PrepareDC(dc);
  // Make sure the back buffer has been repainted properly
  wxRegionIterator upd(GetUpdateRegion());
  while (upd) {
    Repaint_Back_Buffer(wxRect (upd.GetRect()));
    upd++;
  }
  // Blit the back buffer to the screen
  dc.DrawBitmap(_back_buffer, 0, 0, false);
  // If a block is being dragged around, draw it
  if (_dragging != NULL)
    _game->Draw_Block(dc, _dragging);
}

// Update the entire back buffer
void Canvas::Repaint_Back_Buffer()
{
  // Calls the overloaded version of Repaint_Back_Buffer with the
  // entire area of the canvas as argument.
  wxSize canvas_size = GetClientSize();
  Repaint_Back_Buffer(wxRect(0, 0, canvas_size.GetWidth(), canvas_size.GetHeight()));
}

// Update part of the back buffer
void Canvas::Repaint_Back_Buffer(const wxRect& area)
{
  wxBrush back_brush(_back_color, wxSOLID);
  wxMemoryDC mem_dc;
  mem_dc.SelectObject(_back_buffer);
  // Set a clipping area for the DC so we don't screw up stuff outside
  // the area that has to be redrawn.
  mem_dc.SetClippingRegion(area);    

  // If there is no background image, fill the area with the
  // background color.
  if (_background == NULL)
  {
    mem_dc.SetPen(*wxTRANSPARENT_PEN);
    mem_dc.SetBrush(back_brush);
    mem_dc.DrawRectangle(area.x, area.y, area.width, area.height);
  }
  // Else fill it with the background image
  else
  {
    Tile_Back_Image(mem_dc, area);
  }

  // If a game is active, draw the pieces in the area (except for the
  // ones that are being dragged).
  if (_game != NULL)
    _game->Draw(mem_dc, _dragging, area);

  // If the puzzle is completed, draw the 'completed!' message.
  if (_completed)
    Draw_Completed_Message(mem_dc);

  mem_dc.DestroyClippingRegion();
}

void Canvas::Draw_Completed_Message(wxDC& dc)
{
  const int margin = 7, border = 2;
  const wxString completed(wxT("Completed!"));

  // White text on a black rectangle.
  dc.SetBrush(*wxBLACK_BRUSH);
  dc.SetTextForeground(*wxWHITE);
  dc.SetFont(_completed_font);

  // Determine text size and position
  int width, height;
  dc.GetTextExtent(completed, &width, &height);
  int x = GetClientSize().GetWidth() - width - margin;
  // Draw
  dc.DrawRectangle(x - border, margin - border, width + 2 * border, height + 2 * border);
  dc.DrawText(completed, x, margin);
}

// Fill an area with the background image
void Canvas::Tile_Back_Image(wxDC& dc, const wxRect& area)
{
  int width = _background->GetWidth();
  int height = _background->GetHeight();

  // The start values for the loop variables are done like this to
  // make the 'tiling' start at the correct place. It causes some
  // overdraw which is prevented by the clipping rectangle.
  for (int y = area.y - (area.y % height); y < area.height + area.y; y += height)
  {
    for (int x = area.x - (area.x % width); x < area.width + area.x; x += width)
      dc.DrawBitmap(*_background, x, y);
  }
}

// Window gets resized
void Canvas::Resize(wxSizeEvent& event)
{
  wxSize new_size = GetClientSize();

  // Move the pieces that are outsize the window into it.
  if (_game != NULL)
    _game->Constrain_Pieces(new_size.GetWidth(), new_size.GetHeight());

  // Resize back buffer and refresh it
  _back_buffer = wxBitmap(new_size.GetWidth(), new_size.GetHeight());
  Repaint_Back_Buffer();
  Refresh(false);
}

// Left mouse button pressed
void Canvas::Mouse_Down(wxMouseEvent& event)
{
  if (_game != NULL)
  {
    // Find a piece to drag
    _dragging = _game->At_Position(event.GetPosition());

    // If one was found
    if (_dragging != NULL)
    {
      // Set the offset - the difference between the mouse position
      // and the top left corner of the piece
      _drag_offset = _dragging->Position() - event.GetPosition();
      // Update the back buffer (the dragged pieces vanish)
      Repaint_Back_Buffer(_game->Get_Bounding_Box(_dragging));
      // Not refreshing here - until the mouse actually moves nothing
      // changes
    }
  }
}

// Left mouse button released
void Canvas::Mouse_Up(wxMouseEvent& event)
{
  // If something is being dragged
  if (_dragging != NULL)
  {
    // If more than one merge took place we have to redraw everything
    // because everything could have moved.
    bool need_full_redraw = _game->Try_To_Connect(_dragging);
    Piece* released = _dragging;
    _dragging = NULL;

    // Check whether the puzzle is completed
    bool completed = _game->Is_Completed();
    if (completed && !_completed)
    {
      _completed = true;
      need_full_redraw = true;
    }

    // Update the back buffer
    if (need_full_redraw)
      Repaint_Back_Buffer();
    else
      Repaint_Back_Buffer(_game->Get_Bounding_Box(released));
    // Refresh the screen
    Refresh(false);
  }
}

// Mouse moved
void Canvas::Mouse_Move(wxMouseEvent& event)
{
  // If something is being dragged
  if (_dragging != NULL)
  {
    // Constrain new mouse position to window size so pieces do not
    // get dragged out of the window.
    wxSize window = GetClientSize();
    wxPoint mouse = event.GetPosition();
    mouse.x = std::max(mouse.x, 0);
    mouse.y = std::max(mouse.y, 0);
    mouse.x = std::min(mouse.x, window.GetWidth());
    mouse.y = std::min(mouse.y, window.GetHeight());

    // Move the dragged pieces and update the screen
    Move_Dragged_Block(mouse + _drag_offset - _dragging->Position());
  }
  // Else if a clearing area is being specified
  else if (_clear_drag_start.x != -1)
  {
    // Draw a box
    Draw_Clear_Area(event.GetPosition());
  }
}

// Move the dragged pieces and update the screen
void Canvas::Move_Dragged_Block(const wxPoint& distance)
{
  // Find the rectangle that needs to be updated
  wxRect changing(_game->Get_Bounding_Box(_dragging));
  // Grow it a little for the distance moved, so it does not end up
  // outside the buffer.
  Grow_Rectangle(changing, distance);
  
  // First create a buffer with the dragged piece overlaid on the back
  // buffer, and then blit that to the screen. This is to prevent
  // flickering.
  wxBitmap buffer(changing.width, changing.height);

  {
    wxMemoryDC source;
    source.SelectObject(_back_buffer);
    wxMemoryDC target;
    target.SelectObject(buffer);
    // Blit the back buffer into the small buffer
    target.Blit(0, 0, changing.width, changing.height, &source, changing.x, changing.y, wxCOPY);

    // Move the pieces that are being dragged
    _game->Move_Block(_dragging, distance);
    // Draw them to the buffer
    _game->Draw_Block(target, _dragging, wxPoint(-changing.x, -changing.y));
  }

  // Finally, draw the buffer to the screen
  wxClientDC dc(this);
  dc.DrawBitmap(buffer, changing.x, changing.y, false);
}

// Set a background color
void Canvas::Set_Background(const wxColor& color)
{
  _back_color = color;
  // Remove the background image
  delete _background;
  _background = NULL;
  // Update screen
  Repaint_Back_Buffer();
  Refresh(false);
}

// Set a background image
void Canvas::Set_Background(const wxImage& image)
{
  delete _background;
  _background = new wxBitmap(image);
  Repaint_Back_Buffer();
  Refresh(false);
}

// Right mouse button pressed
void Canvas::Start_Clear_Area(wxMouseEvent& event)
{
  // When this is set to something else than {-1, -1}, it means a
  // clear area is being specified.
  _clear_drag_start = event.GetPosition();
}

// Right mouse button released
void Canvas::Clear_Area(wxMouseEvent& event)
{
  // If _clear_drag_start is valid
  if (_clear_drag_start.x != -1)
  {
    // If there's a game
    if (_game != NULL)
    {
      wxSize size = GetClientSize();
      wxPoint mouse = event.GetPosition();
      // Find the rectangle that was specified
      wxPoint top_left(std::min(_clear_drag_start.x, mouse.x), std::min(_clear_drag_start.y, mouse.y));
      wxPoint bottom_right(std::max(_clear_drag_start.x, mouse.x), std::max(_clear_drag_start.y, mouse.y));

      // Move pieces out of that area
      _game->Clear_Area(wxRect(top_left.x, top_left.y, bottom_right.x - top_left.x, bottom_right.y - top_left.y),
                        wxPoint(size.GetWidth(), size.GetHeight()));
      // Refresh buffer
      Repaint_Back_Buffer();
    }

    // No longer dragging
    _clear_drag_start = wxPoint(-1, -1);
    // Repaint (even if nothing was cleared, the rectangle has to be
    // removed)
    Refresh(false);
  }
}

// Draw a rectangle over the area that is being selected to clear
void Canvas::Draw_Clear_Area(const wxPoint& mouse)
{
  // Brush with diagonal stripes
  wxBrush fill_brush(*wxWHITE, wxFDIAGONAL_HATCH);
  wxClientDC dc(this);
  // Blit back buffer to screen
  dc.DrawBitmap(_back_buffer, 0, 0, false);
  dc.SetBrush(fill_brush);
  dc.SetPen(*wxWHITE_PEN);
  // Add rectangle
  dc.DrawRectangle(mouse.x, mouse.y, _clear_drag_start.x - mouse.x, _clear_drag_start.y - mouse.y);
}
