/* Copyright (c) 2004 Marijn Haverbeke
 *
 * This software is provided 'as-is', without any express or implied
 * warranty. In no event will the authors be held liable for any
 * damages arising from the use of this software.
 *
 * Permission is granted to anyone to use this software for any
 * purpose, including commercial applications, and to alter it and
 * redistribute it freely, subject to the following restrictions:
 *
 * 1. The origin of this software must not be misrepresented; you must
 *    not claim that you wrote the original software. If you use this
 *    software in a product, an acknowledgment in the product
 *    documentation would be appreciated but is not required.
 *
 * 2. Altered source versions must be plainly marked as such, and must
 *    not be misrepresented as being the original software.
 *
 * 3. This notice may not be removed or altered from any source
 *    distribution.
 *
 * Marijn Haverbeke
 * marijn(at)haverbeke.nl
 */

// Cutter generates the shapes for the puzzle and adds masks to the
// pieces' images to make them look right.

#include "cutter.hpp"
#include "jigsaw.hpp"

namespace{
  // There are 8 side shapes defined (1 contains the straight sides on
  // the edge of the puzzle)
  const size_t side_styles = 8;
  // An edge mask is as high as a piece
  const size_t mask_height = piece_size;
  // And twice the border size wide
  const size_t mask_width = 2 * border_size;

  // The data for the shapes, defined on the bottom of this file.
  extern const unsigned char* mask_data_p;

  // A mask iterator is used to go through the data for a mask and
  // return whether it is masked for every pixel.
  class Mask_Iterator
  {
  private:
    // Because the data is saved in 1-bit-per-pixel format, some
    // bitwise magic is necessary.
    static const size_t byte_size = 8;
    static const size_t mask_size = (mask_height * mask_width) / byte_size;
    
  public:
    Mask_Iterator(size_t style)
        // Pointer going over the bytes
      : _byte(mask_data_p + style * mask_size),
        // The current bit
        _bit(0)
    {}

    Mask_Iterator& operator++()
    {
      // Go to the next bit
      ++_bit;
      // If that was the last bit, go to the next byte
      if (_bit == 8)
      {
        ++_byte;
        _bit = 0;
      }
      return *this;
    }
    bool operator*()
    {
      // Extract the correct bit
      return ((*_byte) & (1 << _bit)) == 0;
    }

  private:
    const unsigned char* _byte;
    size_t _bit;
  };

  // Convenient random numbers
  inline size_t Random_Int(size_t max)
  {
    return std::rand() % max;
  }
  inline bool Random_Bool()
  {
    return Random_Int(2) == 0;
  }
}

// Generates the shapes for the puzzle
Jigsaw_Cutter::Jigsaw_Cutter(size_t width, size_t height)
    // Vector with width * height shapes
  : _shapes(width * height),
    _width(width),
    _height(height)
{
  // For every piece
  for (size_t y = 0; y != _height; ++y)
  {
    for (size_t x = 0; x != _width; ++x)
    {
      size_t index = x + y * _width;
      // If this is not the bottom row
      if (y != _height - 1)
      {
        // Make up a shape (and whether to reverse it) - the -1+1
        // stuff is to prevent choosing the 'straight' shape which is
        // shape 0.
        size_t shape = Random_Int(side_styles - 1) + 1;
        bool reverse = Random_Bool();

        // Assign this piece and the one below it that shape.
        _shapes[index].bottom.id = shape;
        _shapes[index].bottom.reverse = reverse;
        _shapes[index + _width].top.id = shape;
        _shapes[index + _width].top.reverse = reverse;
      }

      // If this is not the rightmost column
      if (x != _width - 1)
      {
        size_t shape = Random_Int(side_styles - 1) + 1;
        bool reverse = Random_Bool();

        // Give this piece and the one to the right a shape
        _shapes[index].right.id = shape;
        _shapes[index].right.reverse = reverse;
        _shapes[index + 1].left.id = shape;
        _shapes[index + 1].left.reverse = reverse;
      }
      // Note that the shapes of the edges are not set here, they are
      // set correctly by Piece_Shape's default constructor.
    }
  }
}

// Give a piece the correct mask
void Jigsaw_Cutter::Cut_Piece(Piece* piece)
{
  wxBitmap& bitmap = piece->Bitmap();
  // A monochome bitmap to hold the mask
  wxBitmap mask(bitmap.GetWidth(), bitmap.GetHeight(), 1);
  // The piece's place in the puzzle
  wxPoint place = piece->Place();
  // The piece's shape
  Piece_Shape& shape = _shapes[place.x + place.y * _width];

  // These brackets are very necessary, if you use SetMask when the DC
  // is not destructed yet, it does not work properly. (Took me a
  // while to figure out, I can assure you)
  {
    wxMemoryDC dc;
    dc.SelectObject(mask);
    // Make the bitmap white
    dc.SetBrush(*wxWHITE_BRUSH);
    dc.SetPen(*wxWHITE_PEN);
    dc.DrawRectangle(0, 0, picture_size, picture_size);

    // Cut out some parts with black    
    dc.SetBrush(*wxBLACK_BRUSH);
    dc.SetPen(*wxBLACK_PEN);
    // The piece_border x piece_border squares in the corner
    Cut_Corners(dc);
    // And the sides
    Cut_Side_Vertical(dc, shape, true);
    Cut_Side_Vertical(dc, shape, false);
    Cut_Side_Horizontal(dc, shape, true);
    Cut_Side_Horizontal(dc, shape, false);
  }

  // Apply mask to piece's bitmap
  bitmap.SetMask(new wxMask(mask));
}

// Black out the corners of the mask - these are always hidden
void Jigsaw_Cutter::Cut_Corners(wxDC& mask)
{
  size_t d = picture_size - border_size;
  mask.DrawRectangle(0, 0, border_size, border_size);
  mask.DrawRectangle(0, d, border_size, border_size);
  mask.DrawRectangle(d, 0, border_size, border_size);
  mask.DrawRectangle(d, d, border_size, border_size);
}

// Cut a vertical side into a shape
void Jigsaw_Cutter::Cut_Side_Vertical(wxDC& mask, Piece_Shape& shape, bool left)
{
  // Is this shape flipped?
  bool reverse = (left ? shape.left.reverse : shape.right.reverse);
  // What value means the pixel is masked?
  bool on = (left == reverse);
  // Where to draw on the mask
  size_t offset = (left ? 0 : piece_size);

  // This will go through the pixels in the shape from top-left to
  // bottom-right, one row at a time.
  Mask_Iterator bits(left ? shape.left.id : shape.right.id);
  // I might have gone overboard a little on the optimizing here,
  // making two separate loops like this. Creating the pieces is quite
  // slow, so I wanted to fix that. Did not work very well. Anyway,
  // the first loop fills the mask left-to-right and the second
  // right-to-left.
  if (!reverse)
  {
    size_t max_x = mask_width + offset;
    for (size_t y = border_size; y != mask_height + border_size; ++y)
      for (size_t x = offset; x != max_x; ++x)
      {
        if (*bits == on)
          mask.DrawPoint(x, y);
        ++bits;
      }
  }
  else
  {
    size_t min_x = offset - 1;
    size_t start_x = offset + 2 * border_size - 1;
    for (size_t y = border_size; y != mask_height + border_size; ++y)
      for (size_t x = start_x; x != min_x; --x)
      {
        if (*bits == on)
          mask.DrawPoint(x, y);
        ++bits;
      }
  }
}

// Cut a horizontal side into a shape - see Cut_Side_Vertical
void Jigsaw_Cutter::Cut_Side_Horizontal(wxDC& mask, Piece_Shape& shape, bool top)
{
  bool reverse = (top ? shape.top.reverse : shape.bottom.reverse);
  bool on = (top == reverse);
  size_t offset = (top ? 0 : piece_size);

  Mask_Iterator bits(top ? shape.top.id : shape.bottom.id);
  if (!reverse)
  {
    size_t max_y = mask_width + offset;
    for (size_t x = border_size; x != mask_height + border_size; ++x)
      for (size_t y = offset; y != max_y; ++y)
      {
        if (*bits == on)
          mask.DrawPoint(x, y);
        ++bits;
      }
  }
  else
  {
    size_t min_y = offset - 1;
    size_t start_y = offset + 2 * border_size - 1;
    for (size_t x = border_size; x != mask_height + border_size; ++x)
      for (size_t y = start_y; y != min_y; --y)
      {
        if (*bits == on)
          mask.DrawPoint(x, y);
        ++bits;
      }
  }
}

namespace{
  // The data of the shapes. 1 bit is one pixel, there are 8 shape
  // blocks in there. I just wrote a little tool to extract data from
  // BMP files and put it in this format.
  const unsigned char mask_data[] = {
    255, 3, 240, 63, 0, 255, 3, 240, 63, 0, 255, 3, 240, 63, 0, 255, 3, 240, 63, 0, 255, 3, 240, 63,
    0, 255, 3, 240, 63, 0, 255, 3, 240, 63, 0, 255, 3, 240, 63, 0, 255, 3, 240, 63, 0, 255, 3, 240, 63,
    0, 255, 3, 240, 63, 0, 255, 3, 240, 63, 0, 255, 3, 240, 63, 0, 255, 3, 240, 63, 0, 255, 3, 240, 63,
    0, 255, 3, 240, 63, 0, 255, 3, 240, 63, 0, 255, 3, 240, 63, 0, 255, 3, 240, 63, 0, 255, 3, 240, 63,
    0, 255, 3, 240, 63, 0, 255, 3, 240, 63, 0, 255, 3, 240, 63, 0, 255, 3, 240, 63, 0, 255, 3, 240, 63,
    0, 255, 3, 240, 63, 0, 255, 1, 240, 31, 0, 255, 0, 240, 15, 0, 127, 0, 240, 7, 0, 63, 0, 240, 3, 0,
    31, 0, 240, 1, 0, 15, 0, 240, 0, 0, 15, 0, 112, 0, 0, 7, 0, 112, 0, 0, 7, 0, 112, 112, 0, 135, 15,
    240, 252, 0, 255, 31, 240, 255, 1, 255, 31, 240, 255, 1, 255, 31, 240, 255, 0, 207, 15, 112, 248,
    0, 7, 7, 112, 0, 0, 7, 0, 112, 0, 0, 7, 0, 240, 0, 0, 15, 0, 240, 0, 0, 31, 0, 240, 1, 0, 63, 0,
    240, 3, 0, 127, 0, 240, 7, 0, 255, 0, 240, 15, 0, 255, 1, 240, 31, 0, 255, 3, 240, 63, 0, 255, 3,
    240, 31, 0, 255, 0, 240, 7, 0, 63, 0, 240, 3, 0, 31, 0, 240, 1, 0, 15, 0, 240, 0, 0, 7, 0, 112, 0,
    0, 7, 0, 48, 0, 0, 3, 0, 48, 0, 0, 3, 0, 16, 0, 0, 1, 0, 16, 0, 0, 1, 0, 16, 0, 7, 3, 248, 48, 128,
    31, 3, 248, 113, 128, 63, 7, 248, 243, 128, 63, 31, 252, 247, 227, 127, 255, 255, 247, 255, 127,
    255, 255, 247, 255, 63, 255, 255, 243, 255, 63, 255, 255, 243, 255, 31, 255, 255, 241, 255, 31,
    255, 255, 240, 255, 15, 255, 127, 240, 255, 7, 255, 63, 240, 255, 3, 255, 31, 240, 255, 0, 255, 7,
    240, 63, 0, 255, 3, 240, 63, 0, 255, 3, 240, 127, 0, 255, 7, 240, 127, 0, 255, 7, 240, 255, 0, 255,
    15, 240, 255, 0, 255, 15, 240, 255, 1, 255, 31, 240, 255, 1, 255, 31, 240, 255, 1, 255, 31, 240,
    255, 1, 255, 31, 240, 255, 1, 255, 15, 240, 255, 0, 63, 7, 240, 96, 0, 7, 4, 48, 0, 0, 3, 0, 16, 0,
    0, 1, 0, 16, 0, 0, 1, 0, 16, 0, 0, 3, 0, 48, 0, 0, 7, 4, 240, 96, 0, 63, 7, 240, 127, 0, 255, 7,
    240, 63, 0, 255, 3, 240, 63, 0, 255, 3, 240, 63, 0, 255, 3, 240, 63, 0, 255, 3, 240, 63, 0, 255, 3,
    240, 63, 0, 255, 3, 240, 63, 0, 255, 3, 240, 63, 0, 255, 3, 240, 63, 0, 255, 3, 240, 63, 0, 255, 3,
    240, 63, 0, 255, 3, 240, 63, 0, 255, 3, 240, 63, 6, 255, 243, 241, 255, 63, 255, 255, 247, 255, 127,
    255, 255, 255, 255, 255, 255, 255, 255, 255, 127, 255, 255, 247, 255, 63, 255, 243, 241, 63, 6, 255,
    3, 240, 63, 0, 159, 3, 112, 48, 0, 3, 2, 16, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 16, 0, 0, 3,
    2, 112, 48, 0, 159, 3, 240, 63, 0, 255, 3, 240, 63, 0, 255, 3, 240, 63, 0, 255, 3, 240, 63, 0, 255,
    3, 240, 63, 0, 255, 3, 240, 63, 0, 255, 1, 240, 31, 0, 255, 0, 240, 15, 0, 127, 0, 240, 7, 0, 63, 0,
    240, 3, 0, 31, 0, 240, 1, 0, 15, 0, 240, 0, 0, 7, 0, 112, 192, 1, 3, 127, 48, 252, 15, 227, 255, 49,
    255, 31, 255, 255, 243, 255, 63, 255, 255, 243, 255, 63, 255, 255, 243, 255, 63, 255, 255, 243, 255,
    63, 255, 255, 243, 255, 63, 243, 255, 49, 254, 31, 195, 255, 48, 240, 7, 7, 28, 112, 0, 0, 15, 0,
    240, 0, 0, 31, 0, 240, 1, 0, 63, 0, 240, 3, 0, 127, 0, 240, 7, 0, 255, 0, 240, 15, 0, 255, 1, 240,
    31, 0, 255, 3, 240, 63, 0, 255, 3, 240, 63, 0, 255, 3, 240, 63, 0, 255, 3, 240, 63, 0, 255, 3, 240,
    127, 0, 255, 7, 240, 127, 0, 255, 7, 240, 127, 0, 255, 7, 240, 255, 0, 255, 15, 240, 255, 0, 255,
    15, 240, 255, 0, 255, 31, 240, 255, 1, 63, 31, 240, 225, 1, 31, 30, 240, 192, 1, 15, 0, 240, 0, 0,
    15, 0, 240, 0, 0, 15, 0, 240, 192, 1, 31, 30, 240, 225, 1, 63, 31, 240, 255, 1, 255, 15, 240, 255,
    0, 255, 15, 240, 255, 0, 255, 15, 240, 255, 0, 255, 7, 240, 127, 0, 255, 7, 240, 127, 0, 255, 3,
    240, 63, 0, 255, 3, 240, 63, 0, 255, 3, 240, 63, 0, 255, 3, 240, 63, 0, 255, 7, 240, 127, 0, 255,
    15, 240, 255, 1, 255, 63, 240, 255, 3, 255, 127, 240, 255, 7, 255, 255, 240, 255, 15, 255, 255, 240,
    255, 31, 255, 255, 241, 255, 31, 255, 255, 240, 255, 15, 159, 127, 240, 240, 3, 7, 30, 48, 192, 0,
    3, 0, 16, 0, 0, 1, 0, 16, 0, 0, 1, 0, 48, 0, 0, 3, 12, 112, 224, 1, 15, 63, 240, 249, 7, 255, 255,
    240, 255, 15, 255, 255, 241, 255, 31, 255, 255, 241, 255, 15, 255, 255, 240, 255, 15, 255, 127, 240,
    255, 7, 255, 63, 240, 255, 3, 255, 31, 240, 255, 0, 255, 7, 240, 127, 0, 255, 3, 240, 63, 0
  };
  const unsigned char* mask_data_p = mask_data;
}
