/* Copyright (c) 2004 Marijn Haverbeke
 *
 * This software is provided 'as-is', without any express or implied
 * warranty. In no event will the authors be held liable for any
 * damages arising from the use of this software.
 *
 * Permission is granted to anyone to use this software for any
 * purpose, including commercial applications, and to alter it and
 * redistribute it freely, subject to the following restrictions:
 *
 * 1. The origin of this software must not be misrepresented; you must
 *    not claim that you wrote the original software. If you use this
 *    software in a product, an acknowledgment in the product
 *    documentation would be appreciated but is not required.
 *
 * 2. Altered source versions must be plainly marked as such, and must
 *    not be misrepresented as being the original software.
 *
 * 3. This notice may not be removed or altered from any source
 *    distribution.
 *
 * Marijn Haverbeke
 * marijn(at)haverbeke.nl
 */

// This file implements most of the game logic - the behaviour of the
// puzzle pieces.

#include <algorithm>
#include <wx/progdlg.h>
#include "jigsaw.hpp"
#include "cutter.hpp"
#include "filler.hpp"
#include "main.hpp"

namespace{
  // Convenience function
  inline size_t Random_Int(size_t max)
  {
    return std::rand() % max;
  }

  // Are two pieces in the same block
  inline bool In_Same_Block(Piece* one, Piece* two)
  {
    return (one == two) || (one->Block() != no_block && one->Block() == two->Block());
  }
  // Which brings me to the block system. A jigsaw contains, apart
  // from the linked list with all the pieces, a vector of vectors of
  // pieces. Each of these vectors contains a block. Pieces all have a
  // Block() method that tells in which block they are, or returns
  // no_block if they are still alone.
}

// A Jigsaw object contains all the pieces for a puzzle and organizes
// them into 'blocks' of pieces that have been fitted together.
Jigsaw::Jigsaw(wxImage& image, size_t width, size_t height)
{
  wxApp &the_app = wxGetApp();
  wxProgressDialog* progress = new wxProgressDialog(wxT("New jigsaw"), wxT("Creating puzzle pieces..."), width * height, the_app.GetTopWindow());
  int progress_counter = 0;
  progress->Show();

  wxSafeYield();
  // Utility classes to make the pieces
  Jigsaw_Cutter cutter(width, height);
  Jigsaw_Filler filler(image, width, height);

  // Nice dialog to show how far we are with making the puzzle

  // For every position in the puzzle
  for (size_t by = 0; by != height; ++by)
  {
    for (size_t bx = 0; bx != width; ++bx)
    {
      // Create a piece
      _pieces.push_back(new Piece(bx, by));
      // Paint it and cut it to shape
      filler.Fill_Piece(_pieces.back());
      cutter.Cut_Piece(_pieces.back());

      // Update the progress dialog
      wxSafeYield();
      ++progress_counter;
      progress->Update(progress_counter);
    }
  }

  // Remove progress dialog
  progress->Close();
  progress->Destroy();
}

// Delete the pieces when the object dies
Jigsaw::~Jigsaw()
{
  for (Piece_List::iterator i = _pieces.begin(); i != _pieces.end(); ++i)
    delete *i;
}

// Randomly place the pieces on an area
void Jigsaw::Scatter_Pieces(int width, int height)
{
  int max_x = width - picture_size;
  int max_y = height - picture_size;

  // Moves pieces individually. If this is called after blocks have
  // been formed the whole thing fucks up. That is why I only call it
  // when a new game has just been created.
  for (Piece_List::iterator i = _pieces.begin(); i != _pieces.end(); ++i)
    (*i)->Set_Position(wxPoint(Random_Int(max_x), Random_Int(max_y)));
}

// Constrain pieces to a certain area. Is used when the window is
// resized.
void Jigsaw::Constrain_Pieces(int width, int height)
{
  int max_x = width - picture_size;
  int max_y = height - picture_size;

  // Here pieces are moved a block at a time to preserve 'block integrity'
  for (Piece_List::iterator i = _pieces.begin(); i != _pieces.end(); ++i)
  {
    wxPoint pos = (*i)->Position();
    if (pos.x > max_x || pos.y > max_y)
      Constrain_Block((*i), max_x, max_y);
  }
}

// Moves a block into the window area when it happened to end up
// outside because the window got shrunk.
void Jigsaw::Constrain_Block(Piece* piece, int max_x, int max_y)
{
  // Find lower-right corner of the block
  wxPoint block_max(0, 0);
  for (Block_Iterator block(this, piece); block; ++block)
  {
    wxPoint pos = block->Position();
    block_max.x = std::max(block_max.x, pos.x);
    block_max.y = std::max(block_max.y, pos.y);
  }

  // Move block into window
  wxPoint motion = wxPoint(std::min(max_x, block_max.x), std::min(max_y, block_max.y)) - block_max;
  Move_Block(piece, motion);
}

// Clear an area of the window
void Jigsaw::Clear_Area(const wxRect& area, const wxPoint& max)
{
  // A rather dumb method of preventing infinite loops - if randomly
  // moving a block to a better place fails 50 times we stop.
  const int max_safety_counter = 50;

  // For every piece
  for (Piece_List::iterator i = _pieces.begin(); i != _pieces.end(); ++i)
  {
    int safety_counter = 0;
    // While the piece is in the cleared area
    while ((*i)->Overlaps(area))
    {
      // Move it to a random place in the window
      Move_Block(*i, wxPoint(Random_Int(max.x - picture_size), Random_Int(max.y - picture_size)) - (*i)->Position());

      // If that was the 50th try we give up
      ++safety_counter;
      if (safety_counter > max_safety_counter)
      {
        wxMessageBox(wxT("This window is too small to clear that area."), wxT("Problem"), wxOK);
        return;
      }
    }
  }
}

// Find the bounding box of a block
wxRect Jigsaw::Get_Bounding_Box(Piece* piece)
{
  // Top-left and bottom-right points
  wxPoint from(piece->Position());
  wxPoint to(piece->Position());

  // For every piece in the block
  for (Block_Iterator block(this, piece); block; ++block)
  {
    // Update corners
    wxPoint pos = block->Position();
    from.x = std::min(pos.x, from.x);
    from.y = std::min(pos.y, from.y);
    to.x = std::max(pos.x + picture_size, to.x);
    to.y = std::max(pos.y + picture_size, to.y);
  }
  return wxRect(from.x, from.y, to.x - from.x, to.y - from.y);
}

// Draw all pieces overlapping area except those in the same block as leave_out
void Jigsaw::Draw(wxDC& dc, Piece* leave_out, const wxRect& area)
{
  // For every piece (note that this goes from back to front, so the
  // pieces in front are drawn over the ones in back).
  for (Piece_List::reverse_iterator i = _pieces.rbegin(); i != _pieces.rend(); ++i)
  {
    // If this piece is not the one that must be left out, and is not
    // in the same block either, and overlaps area
    if ((leave_out == NULL || !In_Same_Block(*i, leave_out)) && (*i)->Overlaps(area))
      // We draw it
      (*i)->Draw(dc);
  }
}

// Move a block of pieces
void Jigsaw::Move_Block(Piece* piece, const wxPoint& distance)
{
  for (Block_Iterator block(this, piece); block; ++block)
    block->Set_Position(block->Position() + distance);
}

// Draw a block, offset is used to be able to draw into a small buffer
// instead of a buffer the size of the whole screen. Used to draw the
// block that is being dragged.
void Jigsaw::Draw_Block(wxDC& dc, Piece* piece, const wxPoint& offset)
{
  for (Block_Iterator block(this, piece); block; ++block)
    block->Draw(dc, offset);
}

// Find out which piece is at a position, and bring the block that
// piece belong to to the front of the piece list.
Piece* Jigsaw::At_Position(const wxPoint& point)
{
  Piece* retval = NULL;
  // Find the piece
  for (Piece_List::iterator i = _pieces.begin(); i != _pieces.end(); ++i)
  {
    if ((*i)->Hit_By(point))
    {
      retval = *i;
      break;
    }
  }

  // Move it to front
  if (retval != NULL)
    Move_Block_To_Front(retval);
  return retval;
}

// Check whether the puzzle has been completed
bool Jigsaw::Is_Completed() const
{
  // If there is only one block left and it contains all the pieces,
  // the puzzle is done.
  return (_blocks.size() == 1 && _blocks[0].size() == _pieces.size());
}

// Move all the pieces in a block to the front of the list.
void Jigsaw::Move_Block_To_Front(Piece* piece)
{
  // First remove them from the list, taking care not to invalidate
  // our iterator.
  for (Piece_List::iterator i = _pieces.begin(); i != _pieces.end(); )
  {
    Piece_List::iterator last = i;
    ++i;
    if (In_Same_Block(*last, piece))
      _pieces.erase(last);
  }

  // Then add them to the front
  for (Block_Iterator block(this, piece); block; ++block)
    _pieces.push_front(block.p());
}

// A piece has moved. See if it has to get connected to any blocks.
bool Jigsaw::Try_To_Connect(Piece* moved)
{
  bool done = false;
  int merges = 0;
  // Restart every time a new connection is found
  while (!done)
  {
    // If nothing is found we are done after this iteration
    done = true;

    // For each piece in this block
    for (Block_Iterator block(this, moved); block; ++block)
    {
      // Compare to each piece
      for (Piece_List::iterator i = _pieces.begin(); i != _pieces.end(); ++i)
      {
        // If they are not in the same block, but do connect
        if ((!In_Same_Block(*i, block.p())) && (*i)->Connect_To(block.p()))
        {
          // Merge the blocks
          Merge_Blocks(*i, block.p());
          ++merges;
          // Restart loop
          done = false;
          break;
        }
      }
      // This is to break out of second loop if we want to restart
      if (!done)
        break;
    }
  }
  // If more than one block was merged this is returned so the canvas
  // knows it must repaint everything - a chain reaction of merges
  // could invalidate pieces of the screen outside the bounding box of
  // this block.
  return merges > 1;
}

// Take the pieces from two blocks and put them in one block
void Jigsaw::Merge_Blocks(Piece* one, Piece* two)
{
  // The amount block two has to be moved to be aligned with block one
  wxPoint adjustment = one->Distance_To(two);

  // Four different cases, depending on whether the blocks consist of
  // only one piece. If both are only one piece
  if (one->Block() == no_block && two->Block() == no_block)
  {
    // Move two to match one
    two->Set_Position(two->Position() + adjustment);
    // Make a new block and put them in it.
    int new_block = _blocks.size();
    _blocks.push_back(Piece_Vector());
    Add_To_Block(one, new_block);
    Add_To_Block(two, new_block);
  }
  // If one is a single piece
  else if (one->Block() == no_block)
  {
    // Move all of block two to match one
    Move_Block(two, adjustment);
    // Put one in block two
    Add_To_Block(one, two->Block());
  }
  // If two is a single piece
  else if (two->Block() == no_block)
  {
    // Move it
    two->Set_Position(two->Position() + adjustment);
    // Add it to one
    Add_To_Block(two, one->Block());
  }
  // Both are blocks of more than one piece
  else
  {
    // Remember two's block id
    int old_block = two->Block();
    // Move block two
    Move_Block(two, adjustment);
    // Add all pieces in block to to block one
    for (Block_Iterator block_two(this, two); block_two; ++block_two)
      Add_To_Block(block_two.p(), one->Block());
    // Erase block two
    Erase_Block(old_block);
  }
}

// Erase a block
void Jigsaw::Erase_Block(size_t block_id)
{
  // If this block is not at the end of the block vector
  if (block_id != _blocks.size() - 1)
  {
    // Swap it with the end of the block vector
    _blocks[block_id] = _blocks.back();
    // And adjust the block id's of the members
    for (size_t i = 0; i != _blocks[block_id].size(); ++i)
      _blocks[block_id][i]->Set_Block(block_id);
  }
  // Pop the last block off the vector
  _blocks.pop_back();
}

// Add a piece to a block
void Jigsaw::Add_To_Block(Piece* piece, int block_id)
{
  // Set it's id
  piece->Set_Block(block_id);
  // And put it in the right vector
  _blocks[block_id].push_back(piece);
}

// Create a block iterator
Jigsaw::Block_Iterator::Block_Iterator(Jigsaw* game, Piece* piece)
  : _array(NULL),
    _single(NULL),
    _index(0),
    _size(0)
{
  int block = piece->Block();
  // If the piece is a single
  if (block == no_block)
  {
    // Store the pointer
    _single = piece;
    // Make array point to it
    _array = &_single;
    // We have only one piece
    _size = 1;
  }
  // It is member of an actual block
  else
  {
    // Array points to the block vector
    _array = &(game->_blocks[block][0]);
    _size = game->_blocks[block].size();
  }
}

// ,PIECE

// Construct a piece. It starts at 0,0 with an empty bitmap
Piece::Piece(size_t bx, size_t by)
  : _bx(bx),
    _by(by),
    _x(0),
    _y(0),
    _block(no_block),
    _bitmap(picture_size, picture_size)
{
}

// Blit a piece to a DC
void Piece::Draw(wxDC& dc, const wxPoint& offset)
{
  dc.DrawBitmap(_bitmap, _x + offset.x, _y + offset.y, true);
}

// Check whether a piece was hit by a mouse click
bool Piece::Hit_By(const wxPoint& point) const
{
  // Two rectangles are used to exclude the corners of the piece,
  // which are always transparant. Per-pixel checking would be nice
  // here but I was too lazy (the piece would have to know the type
  // and orientation of all of it's sides (see cutter.cpp) etc).
  return ((point.x >= _x + border_size && point.x < _x + picture_size - border_size && point.y >= _y && point.y < _y + picture_size) ||
          (point.x >= _x && point.x < _x + picture_size && point.y >= _y + border_size && point.y < _y + picture_size - border_size));
}

// Does a piece overlap a rectangle
bool Piece::Overlaps(const wxRect& area) const
{
  return (_x + picture_size > area.x && _x <= area.x + area.width &&
          _y + picture_size > area.y && _y <= area.y + area.height);
}

// Distance another piece has to move to match this piece
wxPoint Piece::Distance_To(Piece* other)
{
  int d_bx = _bx - other->_bx;
  int d_by = _by - other->_by;
  wxPoint proper_distance(other->_x + d_bx * piece_size, other->_y + d_by * piece_size);
  return Position() - proper_distance;
}

// Do two pieces connect?
bool Piece::Connect_To(Piece* other)
{
  // Difference in position on the puzzle
  int d_bx = _bx - other->_bx;
  int d_by = _by - other->_by;
  // If that is 2 or more they do not connect
  if (std::abs(d_bx) > 1 || std::abs(d_by) > 1)
    return false;
  // Has to be 0 on at least one side (connecting at a corner is not
  // enough)
  if (d_bx != 0 && d_by != 0)
    return false;

  // Error in positions
  int x_error = (_x - other->_x) - piece_size * d_bx;
  int y_error = (_y - other->_y) - piece_size * d_by;
  // If it is smaller than snap_distance on both x and y they do
  // connect
  return std::abs(x_error) <= snap_distance && std::abs(y_error) <= snap_distance;
}
