/* Copyright (c) 2004 Marijn Haverbeke
 *
 * This software is provided 'as-is', without any express or implied
 * warranty. In no event will the authors be held liable for any
 * damages arising from the use of this software.
 *
 * Permission is granted to anyone to use this software for any
 * purpose, including commercial applications, and to alter it and
 * redistribute it freely, subject to the following restrictions:
 *
 * 1. The origin of this software must not be misrepresented; you must
 *    not claim that you wrote the original software. If you use this
 *    software in a product, an acknowledgment in the product
 *    documentation would be appreciated but is not required.
 *
 * 2. Altered source versions must be plainly marked as such, and must
 *    not be misrepresented as being the original software.
 *
 * 3. This notice may not be removed or altered from any source
 *    distribution.
 *
 * Marijn Haverbeke
 * marijn(at)haverbeke.nl
 */

#ifndef JIGSAW_HPP
#define JIGSAW_HPP

#include <list>
#include <vector>
#include <wx/wx.h>

// Size of the actual pieces
const int piece_size = 50;
// Size of the image for a piece
const int picture_size = 70;
// Border - the area that is not really the piece but can contain part
// of it's shape
const int border_size = (picture_size - piece_size) / 2;

const int no_block = -1;
const int snap_distance = 6;

// A puzzle piece
class Piece
{
public:
  Piece(size_t bx, size_t by);

  void Draw(wxDC& dc, const wxPoint& offset = wxPoint(0, 0));
  
  bool Hit_By(const wxPoint& point) const;
  bool Overlaps(const wxRect& area) const;
  
  void Set_Position(const wxPoint& point) {_x = point.x; _y = point.y;}
  void Set_Block(size_t block) {_block = block;}
  
  wxBitmap& Bitmap() {return _bitmap;}
  wxPoint Position() const {return wxPoint(_x, _y);}
  int Block() const {return _block;}
  wxPoint Place() const {return wxPoint(_bx, _by);}

  wxPoint Distance_To(Piece* other);
  bool Connect_To(Piece* other);
  
private:
  int _bx;
  int _by;
  int _x;
  int _y;

  int _block;
  wxBitmap _bitmap;
};

// A jigsaw puzzle
class Jigsaw
{
public:
  typedef std::list<Piece*> Piece_List;
  typedef std::vector<Piece*> Piece_Vector;

  Jigsaw(wxImage& image, size_t width, size_t height);
  ~Jigsaw();

  void Scatter_Pieces(int width, int height);
  void Constrain_Pieces(int width, int height);
  void Clear_Area(const wxRect& area, const wxPoint& max);
  wxRect Get_Bounding_Box(Piece* piece);
  
  void Draw(wxDC& dc, Piece* leave_out, const wxRect& area);
  void Move_Block(Piece* piece, const wxPoint& distance);
  void Draw_Block(wxDC& dc, Piece* target, const wxPoint& offset = wxPoint(0, 0));

  bool Try_To_Connect(Piece* moved);
  Piece* At_Position(const wxPoint& point);
  bool Is_Completed() const;

private:
  // Iterate through a block. Used to be able to treat pieces that are
  // in a block and pieces that are in not the same way.
  class Block_Iterator
  {
  public:
    Block_Iterator(Jigsaw* game, Piece* piece);
    Piece& operator*() const {return *_array[_index];}
    Piece* operator->() const {return _array[_index];}

    operator bool() const {return _index < _size;}
    Piece* p() const{return _array[_index];}
    Block_Iterator& operator++() {++_index; return *this;}

  private:
    Piece** _array;
    Piece* _single;
    size_t _index;
    size_t _size;
  };
  friend class Block_Iterator;
    
  void Move_Block_To_Front(Piece* piece);
  void Merge_Blocks(Piece* one, Piece* two);
  void Constrain_Block(Piece* piece, int max_x, int max_y);
  void Erase_Block(size_t block_id);
  void Add_To_Block(Piece* piece, int block_id);
  
  Piece_List _pieces;
  std::vector<Piece_Vector> _blocks;
};

#endif //JIGSAW_HPP
