/* Copyright (c) 2004 Marijn Haverbeke
 *
 * This software is provided 'as-is', without any express or implied
 * warranty. In no event will the authors be held liable for any
 * damages arising from the use of this software.
 *
 * Permission is granted to anyone to use this software for any
 * purpose, including commercial applications, and to alter it and
 * redistribute it freely, subject to the following restrictions:
 *
 * 1. The origin of this software must not be misrepresented; you must
 *    not claim that you wrote the original software. If you use this
 *    software in a product, an acknowledgment in the product
 *    documentation would be appreciated but is not required.
 *
 * 2. Altered source versions must be plainly marked as such, and must
 *    not be misrepresented as being the original software.
 *
 * 3. This notice may not be removed or altered from any source
 *    distribution.
 *
 * Marijn Haverbeke
 * marijn(at)haverbeke.nl
 */

// Main.cpp - main application and frame.

#include <ctime>
#include <cstdlib>
#include <wx/config.h>
#include <wx/dirdlg.h>
#ifdef wxUSE_ABOUTDLG
#include <wx/aboutdlg.h>
#endif
#include <wx/filename.h> 
#define wxUSE_COLOURDLG 1
#include <wx/colordlg.h>
#include <wx/image.h>
#include "main.hpp"
#include "canvas.hpp"
#include "jigsaw.hpp"
#include "ids.hpp"
#include "background.hpp"
// Give us an icon on the following ports...
#if defined(__WXGTK__) || defined(__WXMOTIF__) || defined(__WXMAC__) || defined(__WXMGL__) || defined(__WXX11__)
#include "pixmaps/jigsaw-0.xpm"
#endif

namespace{
  // Some big strings that are defined at the bottom of this file.
  extern const wxChar* license;
  extern const wxChar* help;

  // Ask the user for an image file.
  wxString Get_Image_File(const wxString& dialog_caption)
  {
    const wxString file_extensions = wxT("Image files (.bmp, .gif, .jpg, .png, .pcx)|*.bmp;*.gif;*.jpg;*.png;*.pcx");

    wxConfigBase *pConfig = wxConfigBase::Get();
    wxFileName file = pConfig->Read(wxT("/Puzzle/LastImageFileName"), wxFileName::GetCwd());
#if wxMINOR_VERSION <= 6
    int options = wxOPEN | wxCHANGE_DIR;
#else
    int options = wxFD_OPEN | wxFD_CHANGE_DIR | wxFD_PREVIEW;
#endif
    return wxFileSelector(dialog_caption, file.GetPath(), wxT(""), file_extensions, file_extensions, options);
  }
}

IMPLEMENT_APP(Jigsaw_App)

// Initialize the application
bool Jigsaw_App::OnInit()
{
  // Set up the config
  SetVendorName(wxT("OpenJigsaw"));
  wxConfigBase *pConfig = wxConfigBase::Get();
  pConfig->SetRecordDefaults();
  
  // Create and show the main frame
  Main_Frame* frame = new Main_Frame();
  frame->Show(true);
  SetTopWindow(frame);

  // Initialize image handlers (or wx won't load images)
  wxInitAllImageHandlers();
  // Seed random number generator
  std::srand(std::time(NULL));
  
  return true;
}

int Jigsaw_App::OnExit()
{
    delete wxConfigBase::Set((wxConfigBase *)NULL);
    return 0;
}

// Event table for the main frame, all of them menu actions
BEGIN_EVENT_TABLE(Main_Frame, wxFrame)
  EVT_MENU(wxID_NEW, Main_Frame::New_Game)
  EVT_MENU(wxID_CLOSE, Main_Frame::Close_Game)
  EVT_MENU(wxID_EXIT, Main_Frame::Quit)
  EVT_MENU(id_menu_background_default, Main_Frame::Set_Default_Background)
  EVT_MENU(id_menu_background_color, Main_Frame::Set_Background_Color)
  EVT_MENU(id_menu_background_image, Main_Frame::Set_Background_Image)
  EVT_MENU(wxID_HELP, Main_Frame::Help)
  EVT_MENU(id_menu_license, Main_Frame::License)
  EVT_MENU(wxID_ABOUT, Main_Frame::About)
  END_EVENT_TABLE();

// The main frame does not have a lot of responsibilities, exept for
// housing the menus. Direct manipulation of pieces is handled by the
// canvas.
Main_Frame::Main_Frame()
    // A frame of 800 by 600 pixels, maximized if possible.
  : wxFrame(NULL, id_main, wxT("OpenJigsaw v" VERSION), wxPoint(0, 0), wxSize(800, 600), wxMAXIMIZE | wxDEFAULT_FRAME_STYLE),
    // The canvas handles all the dragging and drawing.
    _canvas(new Canvas(this)),
    _game(NULL)
{

  // Set the window icon.
#ifdef __WINDOWS__
  wxIcon icon(wxT("MAIN_ICON"), wxBITMAP_TYPE_ICO_RESOURCE);
  SetIcon(icon);
#else
  SetIcon(wxICON(jigsaw_icon));
#endif


  // Create the menus
  wxMenu* file_menu = new wxMenu;
  file_menu->Append(wxID_NEW, wxT("&New puzzle"));
  file_menu->Append(wxID_CLOSE, wxT("&Close puzzle"));
  file_menu->AppendSeparator();
  file_menu->Append(wxID_EXIT, wxT("E&xit"));

  wxMenu* background_menu = new wxMenu;
  background_menu->Append(id_menu_background_default, wxT("&Default"));
  background_menu->Append(id_menu_background_color, wxT("Solid &color"));
  background_menu->Append(id_menu_background_image, wxT("&Image"));

  wxMenu* help_menu = new wxMenu;
  help_menu->Append(wxID_HELP, wxT("&Help"));
#ifndef wxUSE_ABOUTDLG
  help_menu->Append(id_menu_license, wxT("&License"));
#endif
  help_menu->AppendSeparator();
  help_menu->Append(wxID_ABOUT, wxT("&About"));
  
  wxMenuBar* menu_bar = new wxMenuBar;
  menu_bar->Append(file_menu, wxT("&File"));
  menu_bar->Append(background_menu, wxT("&Background"));
  menu_bar->Append(help_menu, wxT("&Help"));
  SetMenuBar(menu_bar);

  // Restore the frame's position and size
  wxConfigBase *pConfig = wxConfigBase::Get();
  pConfig->SetPath(wxT("/MainFrame"));
  // The (long int)0 casting is needed or the compiler gets confused and calls
  // the wrong member.
  int x = pConfig->Read(_T("x"), (long int)0),
      y = pConfig->Read(_T("y"), (long int)0),
      w = pConfig->Read(_T("w"), 800),
      h = pConfig->Read(_T("h"), 600);
  Move(x, y);
  SetClientSize(w, h);
  pConfig->SetPath(wxT("/"));
  
  // Give the canvas it's initial background
  _canvas->Set_Background(Get_Default_Background());
}

Main_Frame::~Main_Frame()
{
  // Save the frame's size and position
  wxConfigBase *pConfig = wxConfigBase::Get();
  int x, y, w, h;
  GetClientSize(&w, &h);
  GetPosition(&x, &y);
  pConfig->Write(_T("/MainFrame/x"), (long) x);
  pConfig->Write(_T("/MainFrame/y"), (long) y);
  pConfig->Write(_T("/MainFrame/w"), (long) w);
  pConfig->Write(_T("/MainFrame/h"), (long) h);
   
  delete _game;
}

void Main_Frame::Quit(wxCommandEvent& event)
{
  Close(true);
}

void Main_Frame::Help(wxCommandEvent &WXUNUSED(event))
{
  wxMessageBox(help, wxT("OpenJigsaw Help"), wxOK, this);
}

void Main_Frame::License(wxCommandEvent &WXUNUSED(event))
{
  wxMessageBox(license, wxT("OpenJigsaw License"), wxOK, this);
}

void Main_Frame::About(wxCommandEvent &WXUNUSED(event))
{
#if wxUSE_ABOUTDLG
  wxAboutDialogInfo info;
  info.SetName(wxT("OpenJigsaw"));
  info.SetVersion(wxT("Version " VERSION));
  info.SetDescription(wxT("OpenJigsaw is a jigsaw puzzle game. You can choose an image file and it will cut it into little puzzle pieces, which you can then assemble again."));
  info.SetCopyright(wxT("(C) 2004-2008 Marijn Haverbeke"));
  info.AddDeveloper(wxT("Marijn Haverbeke"));
  info.AddDeveloper(wxT("Gregory McLean"));
  info.SetWebSite(wxT("http://marijn.haverbeke.nl/jigsaw/"), wxT("OpenJigsaw web site"));
  info.SetLicense(license);

  wxAboutBox(info);
#else 
  wxMessageBox(wxT("This is version " VERSION " of OpenJigsaw\nWritten by Marijn Haverbeke\nhttp://marijn.haverbeke.nl"),
               wxT("About OpenJigsaw"), wxOK | wxICON_INFORMATION, this);
#endif  
}

void Main_Frame::Set_Default_Background(wxCommandEvent& event)
{
  _canvas->Set_Background(Get_Default_Background());
}

void Main_Frame::Set_Background_Color(wxCommandEvent& event)
{
  // Query the user for a color using the convenient wx color dialog,
  // and pass it to the canvas if a valid color was given.
  wxColor color = wxGetColourFromUser(this, *wxWHITE);
  if (color.Ok())
    _canvas->Set_Background(color);
}

void Main_Frame::Set_Background_Image(wxCommandEvent& event)
{
  // Ask user for a file, if one is given and it yields a proper image
  // we pass it to the canvas.
  wxString filename = Get_Image_File(wxT("Select background image"));
  if (!filename.empty())
  {
    wxImage image(filename);
    if (image.Ok())
      _canvas->Set_Background(image);
  }
}

void Main_Frame::New_Game(wxCommandEvent& event)
{
  // Get a file
  wxString filename = Get_Image_File(wxT("Select image for this puzzle"));
  if (filename.empty())
    return;

  // See if it is a valid image
  wxImage image(filename);
  if (!image.Ok())
    return;

  // Get a size for this image

  wxPoint size = Select_Puzzle_Size(image.GetWidth(), image.GetHeight());
  if (size.x == 0)
    return;
  // Store the path so we can restore it next time we create a puzzle
  wxConfigBase *pConfig = wxConfigBase::Get();
  pConfig->Write(_T("/Puzzle/LastImageFileName"), filename);

  // Delete the old game and make a new game
  delete _game;
  _game = new Jigsaw(image, size.x, size.y);
  _canvas->Set_Game(_game);
}

// Close the current game
void Main_Frame::Close_Game(wxCommandEvent& event)
{
  delete _game;
  _game = NULL;
  _canvas->Set_Game(NULL);
}

wxPoint Main_Frame::Select_Puzzle_Size(int image_width, int image_height)
{
  // Calculate the height/width ratio for the image
  float ratio = static_cast<float>(image_height) / image_width;

  // Query the user for a puzzle size
    const size_t widths[] = {4, 7, 10, 12, 15, 0};
    wxArrayString choices;
    for (const size_t* width = widths; *width != 0; ++width) {
        size_t height = Compute_Height(*width, ratio);
        size_t px_height =  height * piece_size;
        size_t px_width = (*width) * piece_size;
        wxString choice;
        choice << *width * height << wxT(" pieces (") << (*width) << wxT('x') << height << wxT(')') << wxT(" [") << px_width << wxT('x') << px_height << wxT(']');
        choices.Add(choice);
    }
    int width = 0;
    wxSingleChoiceDialog dialog(this,
                                wxT("Select the number of pieces for this puzzle\n")
                                wxT(""),
                                wxT("Select the piece count"),
                                choices);


    if (dialog.ShowModal() == wxID_OK) {
        width = widths[dialog.GetSelection()];
    }

  return wxPoint(width, Compute_Height(width, ratio));
}

namespace{
  // These actually display quite ugly in a dialog because the
  // newlines should be in other places in a variable-width font. I
  // was too lazy to do something about that.
  const wxChar* license = wxT(
    "   Copyright (c) 2004 Marijn Haverbeke\n"
    "\n"
    "This software is provided 'as-is', without any express or implied\n"
    "warranty. In no event will the authors be held liable for any damages\n"
    "arising from the use of this software.\n"
    "\n"
    "Permission is granted to anyone to use this software for any purpose,\n"
    "including commercial applications, and to alter it and redistribute it\n"
    "freely, subject to the following restrictions:\n"
    "\n"
    "1. The origin of this software must not be misrepresented; you must not\n"
    "     claim that you wrote the original software. If you use this software\n"
    "     in a product, an acknowledgment in the product documentation would\n"
    "     be appreciated but is not required.\n"
    "\n"
    "2. Altered source versions must be plainly marked as such, and must not\n"
    "     be misrepresented as being the original software.\n"
    "\n"
    "3. This notice may not be removed or altered from any source\n"
    "     distribution.\n"
    "\n"
    "Source available at: http://marijn.haverbeke.nl/jigsaw\n"
    "Marijn Haverbeke\n"
    "marijn@haverbeke.nl\n");

  const wxChar* help = wxT(
    "OpenJigsaw is a program that converts images into jigsaw puzzles and "
    "lets you solve them. Choose 'new puzzle' from the 'file' menu to start "
    "on a puzzle. You will be asked to pick an image and a size for the "
    "jigsaw and the pieces will appear."
    "\n\n"
    "Move pieces by dragging them with the left mouse button. You can drag "
    "with the right mouse button to clear pieces off a portion of the "
    "screen. If the background color is too similar to some of the pieces "
    "it can be hard to see them. In that case, use the 'background' menu to "
    "pick another color or image to be used as a background.");
}
    
